# Core imports
from tests.utils import make_random_tree
from tests.utils import LIST_SPCDS
from tests.utils import make_tree_list
from fastfuels_core.crown_profile_models.beta import BetaCrownProfile

# External imports
import pytest
import numpy as np


@pytest.fixture
def model():
    # Suppose that I have a tree with a crown base height of 10m and crown length of 10m
    return BetaCrownProfile(species_code=122, crown_base_height=10, crown_length=10)


# generic way to create a custom model however we like for potential expansion of the tests
@pytest.fixture
def custom_model(scope="module"):
    def _model(sc, cbht, cl):
        return BetaCrownProfile(sc, cbht, cl)

    return _model


# generates a list of all the trees in the species code index
# the returned list is formatted in the particular way of PurvesCrownProfile models
@pytest.fixture()
def beta_tree_list():
    trees = make_tree_list("beta")
    tree_list = []
    for t in trees:
        tree = BetaCrownProfile(
            species_code=t.species_code,
            crown_base_height=t.crown_base_height,
            crown_length=t.crown_length,
        )
        tree_list.append(tree)
    return tree_list


@pytest.fixture()
def vector_test_input():
    return np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            15,
            20,
            25,
            30,
            35,
            40,
            45,
            50,
            55,
            60,
            61,
            -1,
        ]
    )


@pytest.fixture()
def scalar_test_input():
    input_pool = np.array(
        [
            0.0,
            1.0,
            2.0,
            3.0,
            4.0,
            5.0,
            6.0,
            7.0,
            8.0,
            9.0,
            10.0,
            15.0,
            20.0,
            25.0,
            30.0,
            35.0,
            40.0,
            45.0,
            50.0,
            55.0,
            60.0,
            61.0,
            -1.0,
        ]
    )
    return input_pool[np.random.randint(input_pool.size)]


vector_length = 100_000


@pytest.fixture(scope="module")
def list_scpds():
    return LIST_SPCDS


@pytest.fixture(scope="module")
def tree_list():
    return [make_random_tree() for _ in range(vector_length)]


@pytest.fixture(scope="module")
def crown_base_height_vector(tree_list):
    return np.array([t.crown_base_height for t in tree_list])


@pytest.fixture(scope="module")
def crown_length_vector(tree_list):
    return np.array([t.crown_length for t in tree_list])


@pytest.fixture(scope="module")
def spcd_vector(tree_list):
    return np.array([t.species_code for t in tree_list])


class TestGetMaxRadius:

    @pytest.mark.parametrize("spcd", LIST_SPCDS)
    def test_individual_tree(self, spcd):
        """
        Tests the individual tree's crown profile using the species code `spcd`. The
        method generates a random tree based on the species code and evaluates the
        maximum crown radius using the Beta crown profile model. It verifies that
        the calculated maximum crown radius meets expected criteria:
        - The value should be a float (scalar)
        - The value should be greater than 0
        - The value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        """
        tree = make_random_tree(species_code=spcd)
        beta_model = BetaCrownProfile(
            species_code=tree.species_code,
            crown_base_height=tree.crown_base_height,
            crown_length=tree.crown_length,
        )
        max_crown_radius = beta_model.get_max_radius()

        # Value should be a float (scalar)
        assert isinstance(max_crown_radius, float)

        # Value should be greater than 0
        assert max_crown_radius > 0.0

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        # The trunc function is used because in some instances where the radius ==,
        # The max_crown_radius has one more decimal point than get_beta_max_crown_radius, causing test failure
        assert np.trunc(max_crown_radius) <= beta_model.get_beta_max_crown_radius(
            (beta_model.crown_base_height + beta_model.crown_length),
            beta_model.crown_base_height,
            beta_model.a,
            beta_model.b,
            beta_model.c,
            beta_model.beta,
        )

    @pytest.mark.parametrize("library", ["numpy", "pandas"])
    def test_tree_vector_numpy(
        self,
        library,
        tree_list,
        crown_base_height_vector,
        crown_length_vector,
        spcd_vector,
    ):
        """
        Tests the tree vector functionality using the specified library,
        either "numpy" or "pandas". The test creates a `BetaCrownProfile`
        model based on the provided tree attributes and checks the resulting
        maximum crown radius. The test ensures that the maximum crown radius
        is calculated correctly and conforms to expected constraints.
        """
        if library == "numpy":
            beta_model = BetaCrownProfile(
                spcd_vector, crown_base_height_vector, crown_length_vector
            )
        elif library == "pandas":
            import pandas as pd

            tree_population_dict = {
                "species_code": spcd_vector,
                "crown_base_height": crown_base_height_vector,
                "crown_length": crown_length_vector,
            }
            tree_population_df = pd.DataFrame(tree_population_dict)
            beta_model = BetaCrownProfile(
                tree_population_df["species_code"],
                tree_population_df["crown_base_height"],
                tree_population_df["crown_length"],
            )
        else:
            raise ValueError(f"Unknown library: {library}")

        max_crown_radius = beta_model.get_max_radius()

        assert isinstance(max_crown_radius, np.ndarray)
        assert len(max_crown_radius) == vector_length
        assert np.all(max_crown_radius > 0.0)
        assert np.all(
            np.trunc(max_crown_radius)
            <= beta_model.get_beta_max_crown_radius(
                (beta_model.crown_base_height + beta_model.crown_length),
                beta_model.crown_base_height,
                beta_model.a,
                beta_model.b,
                beta_model.c,
                beta_model.beta,
            )
        )


class TestGetRadiusAtHeight:
    @pytest.mark.parametrize("spcd", LIST_SPCDS)
    def test_individual_tree(self, spcd, scalar_test_input, vector_test_input):
        """
        Tests the individual tree's crown profile using the species code `spcd`. The
        method generates a random tree based on the species code and evaluates the
        maximum crown radius using the Beta crown profile model. It verifies that
        the calculated maximum crown radius meets expected criteria:
        - The value should be a float (scalar)
        - The value should be greater than 0
        - The value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        """
        tree = make_random_tree(species_code=spcd)
        beta_model = BetaCrownProfile(
            species_code=tree.species_code,
            crown_base_height=tree.crown_base_height,
            crown_length=tree.crown_length,
        )
        # Test scalar input first
        scalar_crown_radius = beta_model.get_radius_at_height(scalar_test_input)

        # Value should be a float (scalar)
        assert isinstance(scalar_crown_radius, float)

        # Value should be greater than 0
        assert scalar_crown_radius >= 0.0

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        # The trunc function is used because in some instances where the radius ==,
        # The max_crown_radius has one more decimal point than get_beta_max_crown_radius, causing test failure
        assert scalar_crown_radius <= beta_model.get_beta_max_crown_radius(
            (beta_model.crown_base_height + beta_model.crown_length),
            beta_model.crown_base_height,
            beta_model.a,
            beta_model.b,
            beta_model.c,
            beta_model.beta,
        )

        # Now test vector input
        vector_crown_radius = beta_model.get_radius_at_height(vector_test_input)

        # Value should be a float (scalar)
        assert isinstance(vector_crown_radius, np.ndarray)

        # Value should be greater than 0
        assert np.all(vector_crown_radius >= 0.0)

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        # The trunc function is used because in some instances where the radius ==,
        # The max_crown_radius has one more decimal point than get_beta_max_crown_radius, causing test failure
        assert np.all(
            vector_crown_radius
            <= beta_model.get_beta_max_crown_radius(
                (beta_model.crown_base_height + beta_model.crown_length),
                beta_model.crown_base_height,
                beta_model.a,
                beta_model.b,
                beta_model.c,
                beta_model.beta,
            )
        )

    @pytest.mark.parametrize("library", ["numpy", "pandas"])
    def test_tree_vector_numpy(
        self,
        library,
        tree_list,
        crown_base_height_vector,
        crown_length_vector,
        spcd_vector,
        scalar_test_input,
        vector_test_input,
    ):
        """
        Tests the tree vector functionality using the specified library,
        either "numpy" or "pandas". The test creates a `BetaCrownProfile`
        model based on the provided tree attributes and checks the resulting
        maximum crown radius. The test ensures that the maximum crown radius
        is calculated correctly and conforms to expected constraints.
        """
        if library == "numpy":
            beta_model = BetaCrownProfile(
                spcd_vector, crown_base_height_vector, crown_length_vector
            )
        elif library == "pandas":
            import pandas as pd

            tree_population_dict = {
                "species_code": spcd_vector,
                "crown_base_height": crown_base_height_vector,
                "crown_length": crown_length_vector,
            }
            tree_population_df = pd.DataFrame(tree_population_dict)
            beta_model = BetaCrownProfile(
                tree_population_df["species_code"],
                tree_population_df["crown_base_height"],
                tree_population_df["crown_length"],
            )
        else:
            raise ValueError(f"Unknown library: {library}")

        # Test scalar input
        scalar_crown_radius = beta_model.get_radius_at_height(scalar_test_input)

        # Value should be a float (scalar)
        assert isinstance(scalar_crown_radius, np.ndarray)

        # Value should be greater than 0
        assert np.all(scalar_crown_radius >= 0.0)

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        # The trunc function is used because in some instances where the radius ==,
        # The max_crown_radius has one more decimal point than get_beta_max_crown_radius, causing test failure
        assert np.all(
            scalar_crown_radius
            <= beta_model.get_beta_max_crown_radius(
                (beta_model.crown_base_height + beta_model.crown_length),
                beta_model.crown_base_height,
                beta_model.a,
                beta_model.b,
                beta_model.c,
                beta_model.beta,
            )
        )

        # Now test vector input
        vector_crown_radius = beta_model.get_radius_at_height(
            np.atleast_2d(vector_test_input).T
        )

        # Value should be a float (scalar)
        assert isinstance(vector_crown_radius, np.ndarray)

        # Value should be greater than 0
        assert np.all(vector_crown_radius >= 0.0)

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        # The trunc function is used because in some instances where the radius ==,
        # The max_crown_radius has one more decimal point than get_beta_max_crown_radius, causing test failure
        assert np.all(
            vector_crown_radius
            <= beta_model.get_beta_max_crown_radius(
                (beta_model.crown_base_height + beta_model.crown_length),
                beta_model.crown_base_height,
                beta_model.a,
                beta_model.b,
                beta_model.c,
                beta_model.beta,
            )
        )



import json
from importlib.resources import files
import numpy as np

from fastfuels_core.crown_profile_models.abc import CrownProfileModel
from scipy.special import beta
from numpy.typing import NDArray

DATA_PATH = files("fastfuels_core.data")
with open(DATA_PATH / "spgrp_parameters.json", "r") as f:
    SPGRP_PARAMS = json.load(f)
with open(DATA_PATH / "spcd_parameters.json", "r") as f:
    SPCD_PARAMS = json.load(f)
with open(DATA_PATH / "class_parameters.json", "r") as f:
    CLASS_PARAMS = json.load(f)


class BetaCrownProfile(CrownProfileModel):
    """
    A crown profile model based on a beta distribution.
    """

    a: float
    b: float
    c: float
    crown_length: float
    crown_base_height: float

    def __init__(
        self, species_code: int, crown_base_height: float, crown_length: float
    ):
        """
        Initializes a BetaCrownProfile instance.
        We initialize data as np arrays expecting those to be passed in
        If only a float (scalar) is passed in, then the functions return results as scalars
        """
        self.species_code = np.asarray(species_code)
        self.crown_base_height = np.asarray(crown_base_height)
        self.crown_length = np.asarray(crown_length)

        # species_group = SPCD_PARAMS[str(self.species_code)]["SPGRP"]
        species_group = vectorized_species_code_lookup(self.species_code, "SPGRP")
        # self.a = SPGRP_PARAMS[str(species_group)]["BETA_CANOPY_a"]
        self.a = vectorized_species_group_lookup(species_group, "BETA_CANOPY_a")
        # self.b = SPGRP_PARAMS[str(species_group)]["BETA_CANOPY_b"]
        self.b = vectorized_species_group_lookup(species_group, "BETA_CANOPY_b")
        # self.c = SPGRP_PARAMS[str(species_group)]["BETA_CANOPY_c"]
        self.c = vectorized_species_group_lookup(species_group, "BETA_CANOPY_c")
        self.beta = beta(self.a, self.b)

    def get_max_radius(self) -> float | np.ndarray:
        """
        Returns the maximum radius of the crown. This function finds the mode
        of the beta distribution, which is the value of z at which the beta
        distribution is at its max, and then calculates the radius at that
        height.

        Returns a scalar with scalar input
        Returns a vector with vector input
        """
        # Find the mode of the beta distribution. This is the value of z at
        # which the beta distribution is at its max.
        z_max = (self.a - 1) / (self.a + self.b - 2)
        normalized_max_radius = self._get_radius_at_normalized_height(z_max)
        return (
            normalized_max_radius * self.crown_length
            if isinstance(normalized_max_radius * self.crown_length, float)
            else np.asarray(normalized_max_radius * self.crown_length)
        )

    def get_radius_at_height(self, height):
        """
        Returns the radius of the crown at a given height using the beta
        distribution crown model described in equation (3) in Ferrarese et
        al. (2015). Equation (3) gives the radius of the crown at a given
        height as a proportion of the crown length scaled between 0 and 1. To
        get a crown radius in meters, the result is multiplied by the crown
        length.
        """
        normalized_height = self._get_normalized_height(height)
        radius_at_normalized_height = self._get_radius_at_normalized_height(
            normalized_height
        )
        return radius_at_normalized_height * self.crown_length

    def _get_normalized_height(self, height):
        """
        Converts a height (in meters) of the tree crown to a unitless height
        between 0 and 1 representing the proportion of the crown length.
        """
        return (height - self.crown_base_height) / self.crown_length

    def _get_radius_at_normalized_height(self, z):
        """
        Returns the unitless radius of the crown at a given normalized height.
        The radius is scaled between 0 and 1 and represents a proportion of
        the crown length of the tree.

        The radius is calculated using the beta distribution probability
        density function (PDF) at the given normalized height. The PDF of the
        beta distribution uses an additional scaling factor, 'c', for the
        application to crown profiles and is described by equation (3) in
        Ferrarese et al. (2015).

        """
        z = np.asarray(z)
        mask = (z >= 0) & (z <= 1)

        result = np.where(
            mask,
            self.c * z ** (self.a - 1) * (1 - z) ** (self.b - 1) / self.beta,
            0.0,
        )
        result = np.nan_to_num(result, nan=0.0)

        if result.size == 1:
            return result.item()  # Return as a scalar
        else:
            return result  # Return as an array

    def get_beta_radius(self, z, height, crown_base, a, b, c, beta):
        """
        Get radius at an array of z heights using the beta crown profile model.

        Parameters
        ----------
        z : NDArray
            Array of z coordinates of float64 type.
        height : float
            Tree height in meters.
        crown_base : float
            Crown base in meters.
        a : float
            Beta distribution parameter.
        b : float
            Beta distribution parameter.
        c : float
            Beta distribution parameter.
        beta : float
            Beta distribution parameter.

        Returns
        -------
        r : NDarray
            Radius of tree evaluated at z heights.
        """

        if np.all(z < crown_base):
            return 0.0
        if np.all(z > height):
            return 0.0

        # Normalize
        crown_length = height - crown_base
        z = (z - crown_base) / crown_length

        r = (c * z ** (a - 1) * (1 - z) ** (b - 1)) / beta
        r = r * crown_length
        return r

    def get_beta_max_crown_radius(self, height, crown_base, a, b, c, beta):
        """
        Gets the maximum radius of a tree for the beta crown profile model.

        Parameters
        ----------
        height : NDArray
            Array of tree heights in meters.
        crown_base : NDArray
            Array of crown base heights in meters.
        crown_base : NDArray
            Crown base heights in meters.
        a : NDArray
            Array of beta distribution parameters.
        b : NDArray
            Array of beta distribution parameters.
        c : NDArray
            Array of beta distribution parameters.
        beta : NDArray
            Array of beta distribution parameters.

        Returns
        -------
        r_max : NDarray
            Maximum radius of each tree in meters.
        """

        # Normalized height of max radius
        z_max = (a - 1) / (a + b - 2)
        # Un-normalized height of max radius
        z_max = crown_base + z_max * (height - crown_base)
        r_max = self.get_beta_radius(z_max, height, crown_base, a, b, c, beta)
        return r_max


# useful function to look up species code. The function is vectorized for convenience.
def _species_code_lookup(species_code: str | NDArray, parameter: str):
    return SPCD_PARAMS[str(species_code)][parameter]


vectorized_species_code_lookup = np.vectorize(_species_code_lookup)


# useful function to look up species groups. The function is vectorized for convenience.
def _species_group_lookup(species_group: str | NDArray, beta_canopy: str):
    return SPGRP_PARAMS[str(species_group)][beta_canopy]


vectorized_species_group_lookup = np.vectorize(_species_group_lookup)



# Core imports
from tests.utils import make_tree_list
from tests.utils import make_random_tree
from tests.utils import LIST_SPCDS
from fastfuels_core.crown_profile_models.purves import PurvesCrownProfile

# External imports
import pytest
import numpy as np


# generates a list of all the trees in the species code index
# the returned list is formatted in the particular way of PurvesCrownProfile models
@pytest.fixture()
def purves_tree_list():
    trees = make_tree_list("purves")
    tree_list = []
    for t in trees:
        tree = PurvesCrownProfile(
            species_code=t.species_code,
            height=t.height,
            dbh=t.diameter,
            crown_ratio=t.crown_ratio,
        )
        tree_list.append(tree)

    return tree_list


@pytest.fixture()
def vector_test_input():
    return np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            15,
            20,
            25,
            30,
            35,
            40,
            45,
            50,
            55,
            60,
            61,
            -1,
        ]
    )


@pytest.fixture()
def scalar_test_input():
    input_pool = np.array(
        [
            0.0,
            1.0,
            2.0,
            3.0,
            4.0,
            5.0,
            6.0,
            7.0,
            8.0,
            9.0,
            10.0,
            15.0,
            20.0,
            25.0,
            30.0,
            35.0,
            40.0,
            45.0,
            50.0,
            55.0,
            60.0,
            61.0,
            -1.0,
        ]
    )
    return input_pool[np.random.randint(input_pool.size)]


# a test sample model we could use
@pytest.fixture
def _model():
    # Suppose that I have a tree with a height of 10m, dbh of 15cm, and crown ratio of 0.4
    return PurvesCrownProfile(species_code=122, height=10, dbh=15, crown_ratio=0.4)


vector_length = 100_000


@pytest.fixture(scope="module")
def list_scpds():
    return LIST_SPCDS


@pytest.fixture(scope="module")
def tree_list():
    return [make_random_tree() for _ in range(vector_length)]


@pytest.fixture(scope="module")
def dbh_vector(tree_list):
    return np.array([t.diameter for t in tree_list])


@pytest.fixture(scope="module")
def height_vector(tree_list):
    return np.array([t.height for t in tree_list])


@pytest.fixture(scope="module")
def crown_ratio_vector(tree_list):
    return np.array([t.crown_ratio for t in tree_list])


@pytest.fixture(scope="module")
def spcd_vector(tree_list):
    return np.array([t.species_code for t in tree_list])


class TestGetMaxRadius:

    @pytest.mark.parametrize("spcd", LIST_SPCDS)
    def test_individual_tree(self, spcd):
        """
        Tests the individual tree's crown profile using the species code `spcd`. The
        method generates a random tree based on the species code and evaluates the
        maximum crown radius using the Purves crown profile model. It verifies that
        the calculated maximum crown radius meets expected criteria:
        - The value should be a float (scalar)
        - The value should be greater than 0
        - The value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        """
        tree = make_random_tree(species_code=spcd)
        purves_model = PurvesCrownProfile(
            species_code=tree.species_code,
            dbh=tree.diameter,
            height=tree.height,
            crown_ratio=tree.crown_ratio,
        )
        max_crown_radius = purves_model.get_max_radius()

        # Value should be a float (scalar)
        assert isinstance(max_crown_radius, float)

        # Value should be greater than 0
        assert max_crown_radius > 0.0

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        assert max_crown_radius <= purves_model._get_purves_max_crown_radius(
            tree.species_code, tree.diameter
        )

    @pytest.mark.parametrize("library", ["numpy", "pandas"])
    def test_tree_vector_numpy(
        self,
        library,
        tree_list,
        dbh_vector,
        height_vector,
        crown_ratio_vector,
        spcd_vector,
    ):
        """
        Tests the tree vector functionality using the specified library,
        either "numpy" or "pandas". The test creates a `PurvesCrownProfile`
        model based on the provided tree attributes and checks the resulting
        maximum crown radius. The test ensures that the maximum crown radius
        is calculated correctly and conforms to expected constraints.
        """
        if library == "numpy":
            purves_model = PurvesCrownProfile(
                spcd_vector, dbh_vector, height_vector, crown_ratio_vector
            )
        elif library == "pandas":
            import pandas as pd

            tree_population_dict = {
                "species_code": spcd_vector,
                "dbh": dbh_vector,
                "height": height_vector,
                "crown_ratio": crown_ratio_vector,
            }
            tree_population_df = pd.DataFrame(tree_population_dict)
            purves_model = PurvesCrownProfile(
                tree_population_df["species_code"],
                tree_population_df["dbh"],
                tree_population_df["height"],
                tree_population_df["crown_ratio"],
            )
        else:
            raise ValueError(f"Unknown library: {library}")

        max_crown_radius = purves_model.get_max_radius()

        assert isinstance(max_crown_radius, np.ndarray)
        assert len(max_crown_radius) == vector_length
        assert np.all(max_crown_radius > 0.0)
        assert np.all(
            max_crown_radius
            <= np.atleast_2d(
                purves_model._get_purves_max_crown_radius(spcd_vector, dbh_vector)
            ).T
        )


class TestGetRadiusAtHeight:

    @pytest.mark.parametrize("spcd", LIST_SPCDS)
    def test_individual_tree(self, spcd, scalar_test_input, vector_test_input):
        """
        Tests the individual tree's crown profile using the species code `spcd`. The
        method generates a random tree based on the species code and evaluates the
        radius at random heights using the Purves crown profile model. It verifies that
        the calculated radius at height meets expected criteria:
        - The value should be a float (scalar) with scalar input
        - The value should be NDArray (vector) with vector input
        - The value should be greater than 0
        - The value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        """
        tree = make_random_tree(species_code=spcd)
        purves_model = PurvesCrownProfile(
            species_code=tree.species_code,
            dbh=tree.diameter,
            height=tree.height,
            crown_ratio=tree.crown_ratio,
        )

        # Test scalar input first
        scalar_crown_radius = purves_model.get_radius_at_height(scalar_test_input)

        # Value should be a float (scalar)
        assert isinstance(scalar_crown_radius, float)

        # Value should be greater than or equal to 0 (height could be larger or smaller than crown height)
        assert scalar_crown_radius >= 0.0

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        assert scalar_crown_radius <= purves_model._get_purves_max_crown_radius(
            tree.species_code, tree.diameter
        )

        # Now test vector input
        vector_crown_radius = purves_model.get_radius_at_height(vector_test_input)

        # Value should be NDArray (vector)
        assert isinstance(vector_crown_radius, np.ndarray)

        # Value should be greater than or equal to 0 (height could be larger or smaller than crown height)
        assert np.all(vector_crown_radius >= 0.0)

        # Value should be less than or equal to the maximum possible crown radius for a tree with Canopy Ratio of 1
        assert np.all(
            vector_crown_radius
            <= np.atleast_2d(
                purves_model._get_purves_max_crown_radius(
                    tree.species_code, tree.diameter
                ).T
            )
        )

    @pytest.mark.parametrize("library", ["numpy", "pandas"])
    def test_tree_vector_numpy(
        self,
        library,
        tree_list,
        dbh_vector,
        height_vector,
        crown_ratio_vector,
        spcd_vector,
        scalar_test_input,
        vector_test_input,
    ):
        """
        Tests the tree vector functionality using the specified library,
        either "numpy" or "pandas". The test creates a `PurvesCrownProfile`
        model based on the provided tree attributes and checks the resulting
        maximum crown radius. The test ensures that the maximum crown radius
        is calculated correctly and conforms to expected constraints.
        """
        if library == "numpy":
            purves_model = PurvesCrownProfile(
                spcd_vector, dbh_vector, height_vector, crown_ratio_vector
            )
        elif library == "pandas":
            import pandas as pd

            tree_population_dict = {
                "species_code": spcd_vector,
                "dbh": dbh_vector,
                "height": height_vector,
                "crown_ratio": crown_ratio_vector,
            }
            tree_population_df = pd.DataFrame(tree_population_dict)
            purves_model = PurvesCrownProfile(
                tree_population_df["species_code"],
                tree_population_df["dbh"],
                tree_population_df["height"],
                tree_population_df["crown_ratio"],
            )
        else:
            raise ValueError(f"Unknown library: {library}")

        # Test scalar test input
        scalar_crown_radius = purves_model.get_radius_at_height(scalar_test_input)

        assert isinstance(scalar_crown_radius, np.ndarray)
        assert len(scalar_crown_radius) == vector_length
        assert np.all(scalar_crown_radius >= 0.0)
        assert np.all(
            scalar_crown_radius
            <= np.atleast_2d(
                purves_model._get_purves_max_crown_radius(spcd_vector, dbh_vector)
            ).T
        )
        # Test vector test input
        vector_crown_radius = purves_model.get_radius_at_height(vector_test_input)

        assert isinstance(vector_crown_radius, np.ndarray)
        assert len(vector_crown_radius) == vector_length
        assert np.all(vector_crown_radius >= 0.0)
        assert np.all(
            vector_crown_radius
            <= np.atleast_2d(
                purves_model._get_purves_max_crown_radius(spcd_vector, dbh_vector)
            ).T
        )



import numpy as np
import random

from fastfuels_core.fuel_moisture.fosberg import (
    calculate_1hr_fuel_moisture,
    _calculate_reference_fuel_moisture,  # noqa
    _calculate_1hr_fuel_moisture_corrections,  # noqa
    _convert_aspect_to_cardinal,  # noqa
    TEMP_BREAKPOINTS,
    RH_BREAKPOINTS,
    RFM_TABLE,
    TABLE_B,
    TABLE_C,
    TABLE_D,
    ShadingBoolean,
    Aspect,
    TIME_RANGES,
    SLOPE_RANGES,
)

import pytest


class TestCalculateReferenceFuelMoisture:

    @pytest.fixture(scope="class")
    def table_test_cases(self):
        cases = []
        for i in range(len(TEMP_BREAKPOINTS)):
            for j in range(len(RH_BREAKPOINTS)):
                # Get a random temperature in range
                temp_low = TEMP_BREAKPOINTS[i]
                temp_high = TEMP_BREAKPOINTS[min(i + 1, len(TEMP_BREAKPOINTS) - 1)]
                if i == len(TEMP_BREAKPOINTS) - 1:
                    temp = random.uniform(temp_low, temp_low + 20)
                else:
                    temp = random.uniform(temp_low, temp_high)

                # Get a random RH in range
                rh_low = RH_BREAKPOINTS[j]
                rh_high = RH_BREAKPOINTS[min(j + 1, len(RH_BREAKPOINTS) - 1)]
                rh = random.uniform(rh_low, rh_high)

                expected = RFM_TABLE[i][j]

                cases.append((temp, rh, expected, i, j))

        return cases

    def test_table(self, table_test_cases):
        for temp, rh, expected, i, j in table_test_cases:
            result = _calculate_reference_fuel_moisture(temp, rh)
            assert result == expected, (
                f"Failed for temp={temp:.2f}, rh={rh:.2f}. Expected {expected}, but got {result}. "
                f"This corresponds to temperature range index {i} and RH range index {j}."
            )

    @pytest.mark.parametrize(
        "temp, rh, expected",
        [
            (10, 0, 1),  # Minimum temp, minimum RH
            (10, 100, 14),  # Minimum temp, maximum RH
            (109, 0, 1),  # Just below maximum temp range, minimum RH
            (109, 100, 13),  # Just below maximum temp range, maximum RH
            (110, 0, 1),  # Maximum temp range, minimum RH
            (110, 100, 12),  # Maximum temp range, maximum RH
            (200, 50, 7),  # Well above maximum temp range, medium RH
        ],
    )
    def test_boundary_conditions(self, temp, rh, expected):
        assert (
            _calculate_reference_fuel_moisture(temp, rh) == expected
        ), f"Boundary condition failed for temp={temp}, rh={rh}. Expected {expected}, but got {_calculate_reference_fuel_moisture(temp, rh)}"

    def test_invalid_inputs(self):
        with pytest.raises(ValueError):
            _calculate_reference_fuel_moisture(9, 50)
        with pytest.raises(ValueError):
            _calculate_reference_fuel_moisture(70, -1)
        with pytest.raises(ValueError):
            _calculate_reference_fuel_moisture(70, 101)

    def test_array_inputs_outputs(self):
        temps = np.array([[75, 91], [55, 72]])
        rhs = np.array([[20, 70], [40, 10]])
        expected = np.array([[3, 9], [6, 2]])
        result = _calculate_reference_fuel_moisture(temps, rhs)
        assert np.array_equal(
            result, expected
        ), f"Failed for array inputs. Expected {expected}, but got {result}"


class TestConvertAspectToCardinal:
    test_size = 100

    @pytest.mark.parametrize(
        "aspect, expected",
        [(0, 0), (45, 1), (90, 1), (135, 2), (180, 2), (225, 3), (270, 3), (315, 0)],
    )
    def test_edges(self, aspect, expected):
        assert (
            _convert_aspect_to_cardinal(aspect) == expected
        ), f"Failed for edge case {aspect}. Expected {expected}, but got {_convert_aspect_to_cardinal(aspect)}"

    @pytest.mark.parametrize("direction", ["North", "East", "South", "West"])
    def test_directions(self, direction):
        if direction == "North":
            expected = 0
            cases = np.random.uniform(low=360 - 45, high=360 + 45, size=self.test_size)
        elif direction == "East":
            expected = 1
            cases = np.random.uniform(low=90 - 45, high=90 + 45, size=self.test_size)
        elif direction == "South":
            expected = 2
            cases = np.random.uniform(low=180 - 45, high=180 + 45, size=self.test_size)
        else:
            expected = 3
            cases = np.random.uniform(low=270 - 45, high=270 + 45, size=self.test_size)
        cases %= 360
        for case in cases:
            assert (
                _convert_aspect_to_cardinal(case) == expected
            ), f"Failed for {direction} direction in case {case}. Expected {expected}, but got {_convert_aspect_to_cardinal(case)}"

    def test_array(self):
        aspects = np.array([[0, 45, 90, 135], [180, 225, 270, 315]])
        expected = np.array([[0, 1, 1, 2], [2, 3, 3, 0]])
        result = _convert_aspect_to_cardinal(aspects)
        assert np.array_equal(
            result, expected
        ), f"Failed for array inputs. Expected {expected}, but got {result}"


class TestCalculate1HrFuelMoistureCorrections:

    @pytest.mark.parametrize("correction_table", [TABLE_B, TABLE_C, TABLE_D])
    @pytest.mark.parametrize(
        "shading_category", [ShadingBoolean.Unshaded, ShadingBoolean.Shaded]
    )
    @pytest.mark.parametrize(
        "aspect_category", [Aspect.North, Aspect.East, Aspect.South, Aspect.West]
    )
    @pytest.mark.parametrize("slope_range", [SLOPE_RANGES[0], SLOPE_RANGES[1]])
    @pytest.mark.parametrize("time_range", [*TIME_RANGES])
    @pytest.mark.parametrize("relative_elevation", [0, 1, 2])
    def test_scalar_inputs(
        self,
        correction_table,
        shading_category,
        aspect_category,
        slope_range,
        time_range,
        relative_elevation,
    ):
        shading = (
            random.uniform(0, 0.5)
            if shading_category == ShadingBoolean.Unshaded
            else random.uniform(0.5, 1)
        )
        aspect = random.uniform(aspect_category * 90 - 45, aspect_category * 90 + 45)
        slope = (
            random.uniform(0, 30)
            if slope_range == SLOPE_RANGES[0]
            else random.uniform(30, 90)
        )
        lower_time = int(time_range.split("-")[0])
        upper_time = int(time_range.split("-")[1])
        time = int(str(int(random.uniform(lower_time, upper_time))).zfill(4))

        result = _calculate_1hr_fuel_moisture_corrections(
            correction_table, shading, aspect, slope, time, relative_elevation
        )
        assert isinstance(
            result, (int, float)
        ), "Result should be a scalar value"  # assert 0 <= result <= 6, "Result should be between 0 and 6"
        assert (
            correction_table[shading_category][aspect_category][slope_range][
                time_range
            ][relative_elevation]
            == result
        )

    def test_array_inputs(self):
        shading = np.array(
            [
                [ShadingBoolean.Unshaded, ShadingBoolean.Shaded],
                [ShadingBoolean.Unshaded, ShadingBoolean.Shaded],
            ]
        )

        aspect = np.array([[Aspect.North, Aspect.South], [Aspect.East, Aspect.West]])
        slope = np.array([[15, 45], [1, 30]])
        time = 1200
        relative_elevation = 1

        result = _calculate_1hr_fuel_moisture_corrections(
            TABLE_B, shading, aspect, slope, time, relative_elevation
        )
        assert isinstance(result, np.ndarray), "Result should be a numpy array"
        assert result.shape == (2, 2), "Result shape should match input shape"

        expected = np.array(
            [
                [
                    TABLE_B[0][0][SLOPE_RANGES[0]][TIME_RANGES[2]][1],
                    TABLE_B[1][2][SLOPE_RANGES[1]][TIME_RANGES[2]][1],
                ],
                [
                    TABLE_B[0][1][SLOPE_RANGES[0]][TIME_RANGES[2]][1],
                    TABLE_B[1][3][SLOPE_RANGES[0]][TIME_RANGES[2]][1],
                ],
            ]
        )
        assert np.array_equal(
            result, expected
        ), f"Failed for array inputs. Expected {expected}, but got {result}"


class TestCalculate1HRFuelMoisture:
    def test_example_scalar(self):
        temp = 75
        rh = 30
        aspect = 180
        slope = 20
        time = 1200
        month = "June"
        elevation = 1
        shading = 0.2

        result = calculate_1hr_fuel_moisture(
            temp, rh, aspect, slope, time, month, elevation, shading
        )
        expected = 5
        assert (
            result == expected
        ), f"Failed for scalar inputs. Expected {expected}, but got {result}"

    def test_example_array(self):
        temps = np.array([80, 70])
        rhs = np.array([30, 50])
        aspects = np.array([180, 270])
        slope = 20
        time = 1200
        month = "June"
        elevation = 1
        shading = 0.2

        result = calculate_1hr_fuel_moisture(
            temps, rhs, aspects, slope, time, month, elevation, shading
        )

        expected = np.array([5, 7])
        assert np.array_equal(
            result, expected
        ), f"Failed for array inputs. Expected {expected}, but got {result}"



# Core imports
from math import ceil
from typing import Optional
import numpy.random as random

# Internal imports
from fastfuels_core.trees import TreePopulation
from fastfuels_core.crown_profile_models.purves import PurvesCrownProfile
from fastfuels_core.crown_profile_models.beta import BetaCrownProfile

# External imports
import numpy as np
from numba import njit
import dask.array as da
from xarray import DataArray
from numpy.typing import NDArray
from dask_image.ndfilters import uniform_filter


class DiscSampler:
    """
    Poisson disk sampling object that refines the placement of trees based
    on canopy intersection and height distribution.

    Parameters
    ----------
    trees : TreePopulation
        An object representating a tree population.
    height_grid : DataArray
        An estimated canopy height model for the region of interest.
    sample_heights : bool
        Should the height of trees be accounted for when sampling?
    crown_profile_model : str
        The type of crown profile model to use when looking for crown overlap.
    """

    def __init__(
        self,
        trees: TreePopulation,
        height_grid: Optional[DataArray] = None,
        sample_heights=True,
        crown_profile_model="purves",
    ):
        # Get tree properties as an array for use in Numba
        trees = trees.dropna()
        self.trees = TreePopulation(trees)

        # What type of crown profile model to use
        if crown_profile_model not in ["beta", "purves"]:
            raise ValueError(
                "The crown profile model must be one of the following: 'beta' or 'purves'"
            )
        self.crown_profile_model = crown_profile_model
        if self.crown_profile_model == "beta":
            height = trees["HT"].to_numpy()
            crown_ratio = trees["CR"].to_numpy()
            crown_length = height * crown_ratio
            crown_base_height = height - crown_length

            # create BetaCrownProfile
            self.beta_model = BetaCrownProfile(
                trees["SPCD"].to_numpy().astype(str), crown_base_height, crown_length
            )
            # update tree properties
            a = self.beta_model.a
            b = self.beta_model.b
            c = self.beta_model.c
            beta = self.beta_model.beta
            trees["MAX_RADIUS"] = self.beta_model.get_beta_max_radius(
                height,
                crown_base_height,
                a,
                b,
                c,
                beta,
            )
        else:  # create purves model
            self.purves_model = PurvesCrownProfile(
                trees["SPCD"].to_numpy().astype(str),
                trees["DIA"].to_numpy().astype(float),
                trees["HT"].to_numpy().astype(float),
                trees["CR"].to_numpy().astype(float),
            )
            # update tree properties
            trees["MAX_RADIUS"] = self.purves_model._get_purves_max_crown_radius(
                self.purves_model.species_code, self.purves_model.dbh
            )

        # Tree coordinates
        x = trees["X"].to_numpy()
        y = trees["Y"].to_numpy()
        x_min = x.min()
        y_min = y.min()
        x_max = x.max()
        y_max = y.max()
        self.bounds = (x_min, y_min, x_max, y_max)

        # Default grid size if none is provided
        self.dx = 0.816281605977565

        # Restrict the CHM to the TreeMap
        if not (height_grid is None):
            self.dx = abs(height_grid.x.data[1] - height_grid.x.data[0])
            height_grid = height_grid.rio.clip_box(
                minx=x_min, miny=y_min, maxx=x_max, maxy=y_max
            )

        xi, yi = self.map_to_pixel_coords(x, y)

        # If there isn't a height grid, make a dummy height grid
        if height_grid is None:
            self.chm = np.ones((xi.max() + 1, yi.max() + 1)).astype(np.float32)
        else:
            self.height_grid = height_grid
            self.chm = height_grid.data[0][::-1, :].astype(np.float32)

        # No trees starting on the edge
        xi[xi < 1] = 1
        yi[yi < 1] = 1
        xi[xi > self.chm.shape[0] - 1] = self.chm.shape[0] - 1
        yi[yi > self.chm.shape[1] - 1] = self.chm.shape[1] - 1
        self.tree_coords = np.c_[xi, yi]

        # If local height is used in sampling, then compute a z-score of
        # height in a distribution of its neighbors
        if sample_heights:
            mask = np.zeros_like(self.chm)
            mask[self.chm > 0.1] = 1.0
            self.mask = mask

            # Do the computations in dask for speed
            mask_da = da.from_array(self.mask, chunks=(1000, 1000))
            chm_da = da.from_array(self.chm, chunks=(1000, 1000))

            # Window size
            w = 15
            mask_m = uniform_filter(mask_da, w)
            mask_sum = mask_m.compute() * w**2

            # Local Mean
            chm_m = uniform_filter(chm_da, w)
            chm_sum = chm_m.compute() * w**2
            chm_mu = chm_sum / (mask_sum + 1.0)

            # Local standard deviation
            chm_s = uniform_filter((chm_da - chm_m) ** 2 * mask, w)
            chm_s_sum = chm_s.compute() * w**2
            chm_sigma = np.sqrt(chm_s_sum / (mask_sum + 1.0))
            chm_sigma[chm_sigma < 1.0] = 1.0

            # Now compute z-score
            self.z_score = (self.chm - chm_mu) / chm_sigma
        else:
            # Otherwize fabricate a z score field
            self.z_score = np.ones_like(self.chm) * 100.0

    def map_to_pixel_coords(self, x, y):
        """
        Convert map coordinates to pixel indexes on a grid.

        Parameters
        ----------
        x : NDArray
            x map coordinate.
        y : NDArray
            y map coordinate.

        Returns
        -------
        xi : ndarray
            Row indexes on a grid.
        yi : NDArray
            Column indexes on a grid.
        """

        (x_min, y_min, x_max, y_max) = self.bounds
        dx = self.dx
        xi = y - y_min
        yi = x - x_min
        xi = np.rint(xi / dx).astype(int)
        yi = np.rint(yi / dx).astype(int)
        return xi, yi

    def pixel_to_map_coords(self, xi, yi):
        """
        Convert pixel indexes on a grid to map coordinates.

        Parameters
        ----------
        xi : ndarray
            Row indexes on a grid.
        yi : NDArray
            Column indexes on a grid.

        Returns
        -------
        x : NDArray
            x map coordinate.
        y : NDArray
            y map coordinate.

        """

        dx = self.dx
        (x_min, y_min, x_max, y_max) = self.bounds
        x = yi * dx + x_min
        y = xi * dx + y_min
        return x, y

    def sample(
        self,
        max_dist: Optional[NDArray] = None,
        tree_scale: Optional[NDArray] = None,
        height_error: Optional[NDArray] = None,
    ) -> TreePopulation:
        """
        Performs Poisson disc sampling to refine placement of trees in the tree population.
        The main criteria are:
        1) Place trees on areas where the CHM > 1 m, if provided.
        2) Avoid excessive crown overlap between trees.
        3) Place tallest trees in a plot in places where the CHM is higher.
        If no CHM has been provided, then sampling just tries to reduce crown overlap.

        Parameters
        ----------
        max_dist (optional): NDArray
            Maximum distance that a tree can move from its original position.
            Can change with iteration. If not specified, use a fixed 30 m per iteration.
        tree_scale (optional): NDArray
            On each sampling iteration, the radius of each tree is multiplied by this factor.
        height_error (optional): NDArray
            A tree can be placed on a pixel if it has a smaller z score than the z score of the CHM
            + height_error. Hence, larger height_error means less stringent placement based on height.


        Returns
        -------
        TreePopulation
            The population of trees with refined x, y coordinates.
        """

        if max_dist is None:
            # Number of sampling iterations
            iterations = 1000
            # Maximum distance a tree can be from its original spot per iteration
            max_dist = np.ones(iterations) * 30.0
            # Paramter that reduces the intersection radius of trees with iteratations
            tree_scale = np.linspace(1.0, 0.0, iterations) ** 2
            # Height error per iteration
            height_error = np.linspace(0.0, 5.0, iterations)

        chm = self.chm
        z_score = self.z_score
        tree_grid = np.zeros_like(self.chm, dtype=np.int64)
        tree_coords = self.tree_coords
        dx = self.dx
        tree_props = self.tree_props

        # Do the disk sampling
        tree_grid, trees_to_place = sample_trees(
            chm,
            z_score,
            tree_grid,
            tree_coords,
            tree_props,
            dx,
            tree_scale,
            max_dist,
            height_error,
        )
        # Get the pixel coordinates of trees and convert to map coords
        coords = np.argwhere(tree_grid >= 1)
        x, y = self.pixel_to_map_coords(coords[:, 0], coords[:, 1])

        # Set new coordinates of the tree population
        trees = self.trees
        trees["X"] = x
        trees["Y"] = y

        return trees


@njit
def propose_point(min_dist: float, max_dist: float, dx=1.0):
    """
    Utility function that proposes a point between a minimum and maximum
    distance from the origin.

    Parameters
    ----------
    min_radius : float
        Minimum distance of proposed point from origin.
    max_radius: float
        Maximum distance of proposed point from origin.
    dx : float
        Resolution of grid cell.

    Returns
    -------
    (int, int)
        Relative grid offsets of proposed point.
    """

    r = min_dist + random.random() * (max_dist - min_dist)
    theta = 2.0 * np.pi * random.random()
    di = r * np.cos(theta)
    dj = r * np.sin(theta)
    di = int(di / dx)
    dj = int(dj / dx)

    return (di, dj)


@njit
def propose_index(
    i: int,
    j: int,
    ni: int,
    nj: int,
    dx=1.0,
    min_radius=1.5,
    max_radius=10.0,
):
    """
    Simple utility function that takes in indices on a grid and proposes new indices of a pixel on the grid
    within an annulus. The proposed point is  constrained to be on the grid.

    Parameters
    ----------
    i : int
        Row index of initial point.
    j : int
        Column index of initial point
    ni : int
        Number of rows on grid.
    nj : int
        Number of columns on grid.
    dx : float
        Resolution of grid cell.
    min_radius : float
        Minimum distance of proposed point from original points.
    max_radius: float
        Maximum distance of proposed point from original point.

    Returns
    -------
    (int, int)
        Grid indices of proposed point.
    """

    # Propose new index
    di, dj = propose_point(min_radius, max_radius, dx)
    i1 = i + di
    j1 = j + dj

    # Make sure it's in bounds
    i1 = max(i1, 1)
    i1 = min(i1, ni - 1)
    j1 = max(j1, 1)
    j1 = min(j1, nj - 1)

    return (i1, j1)


@njit
def intersects(
    distance: float,
    props0: NDArray,
    props1: NDArray,
    scale=1.0,
    crown_profile_model="purves",
):
    """
    A generic wrapper function for detecting crown intersection for different crown profile models.

    Parameters
    ----------
    distance : float
        Distance between two trees
    props0 : NDArray
        Geometric parameters of first tree.
    props0 : NDArray
        Geometric parameters of second tree tree.
    scale : float
        The radius of each tree is multiplied by this scaling factor.
    crown_profile_model: str
        The type of crown profile model to use.

    Returns
    -------
    bool
        Do the two trees with radii multiplied by scale intersect?
    """

    if crown_profile_model == "beta":
        intersects = detect_intersection_beta(distance, props0, props1, scale)
    else:
        intersects = detect_intersection_purves(distance, props0, props1, scale)

    return intersects


@njit
def detect_intersection_purves(
    distance: float, props0: NDArray, props1: NDArray, scale=1.0
):
    """
    Detects intersections between trees using the Purves crown profile model.

    Parameters
    ----------
    distance : float
        Distance between two trees
    props0 : NDArray
        Geometric parameters of first tree.
    props0 : NDArray
        Geometric parameters of second tree tree.
    scale : float
        The radius of each tree is multiplied by this scaling factor.
    crown_profile_model: str
        The type of crown profile model to use.

    Returns
    -------
    bool
        Do the two trees with radii multiplied by scale intersect?
    """

    height0 = props0[0]
    crown_ratio0 = props0[1]
    max_radius0 = props0[2]
    shape0 = props0[3]
    crown_base0 = height0 - crown_ratio0 * height0

    height1 = props1[0]
    crown_ratio1 = props1[1]
    max_radius1 = props1[2]
    shape1 = props1[3]
    crown_base1 = height1 - crown_ratio1 * height1

    cb_max = max(crown_base0, crown_base1)
    r0_max = scale * PurvesCrownProfile._get_purves_radius(
        cb_max, height0, crown_base0, max_radius0, shape0
    )
    r1_max = scale * PurvesCrownProfile._get_purves_radius(
        cb_max, height1, crown_base1, max_radius1, shape1
    )

    return r0_max + r1_max >= distance


@njit
def detect_intersection_beta(
    distance: float, props0: NDArray, props1: NDArray, scale=1.0
):
    """
    Detects intersections between trees using the beta crown profile model. Slightly less
    efficient than Purves model implementation.

    Parameters
    ----------
    distance : float
        Distance between two trees
    props0 : NDArray
        Geometric parameters of first tree.
    props0 : NDArray
        Geometric parameters of second tree tree.
    scale : float
        The radius of each tree is multiplied by this scaling factor.
    crown_profile_model: str
        The type of crown profile model to use.

    Returns
    -------
    bool
        Do the two trees with radii multiplied by scale intersect?
    """

    height0 = props0[0]
    crown_ratio0 = props0[1]
    a0 = props0[3]
    b0 = props0[4]
    c0 = props0[5]
    beta0 = props0[6]
    crown_base0 = height0 - crown_ratio0 * height0

    height1 = props1[0]
    crown_ratio1 = props1[1]
    a1 = props1[3]
    b1 = props1[4]
    c1 = props1[5]
    beta1 = props1[6]
    crown_base1 = height1 - crown_ratio1 * height1

    z0 = max(crown_base0, crown_base1)
    z1 = min(height0, height1)
    dz = (z1 - z0) / 8.0
    # print(z0, z1, dz)

    z = z0
    while z <= z1:
        r0 = scale * BetaCrownProfile.get_beta_radius(
            z, height0, crown_base0, a0, b0, c0, beta0
        )
        r1 = scale * BetaCrownProfile.get_beta_radius(
            z, height1, crown_base1, a1, b1, c1, beta1
        )
        if r0 + r1 >= distance:
            return True
        z += dz

    return False


@njit
def accept(
    chm: NDArray,
    z_score: NDArray,
    tree_grid: NDArray,
    tree_props: NDArray,
    tree_index: int,
    i: int,
    j: int,
    dx=1.0,
    min_radius=1.0,
    tree_scale=1.0,
    height_error=0.0,
    crown_profile_model="purves",
):
    """
    Essentially the heart of the disc sampling routine. Determines whether or not
    to accept the proposed placement for a tree on a grid.

    Parameters
    ----------
    chm: NDArray
        Crown height model (CHM).
    z_score: NDArray
        A z_score for each CHM pixel representing its height compared to its neighbors.
    tree_grid: NDArray
        An array of the same size as the CHM used to mark tree locations.
    tree_props: NDArray
        The geometric parameters of each tree in a tree population.
    tree_index : int
        A tree index that can be used to get tree properties int the tree_props array.
    i : int
        Proposed row to place tree.
    j : int
        Proposed column to palace tree.
    min_radius : float
        A strict minimum distance between trees. Usually just the grid size.
    tree_scale : float
        Rescale tree radii when detecting crown overlap.
    height_error : float
        A tree can be placed on a pixel if it has a smaller z score than the z score of the CHM
        + height_error. Hence, larger height_error means less stringent placement based on height.
    crown_profile_model : str
        The crown profile model to use.

    Returns
    -------
    bool
        Returns tree if we accept the proposed tree placement, false otherwise.
    """

    # Only consider a point where the normalized CHM is above a given threshold
    valid = chm[i, j] >= 1.0 and tree_grid[i, j] == 0
    if not valid:
        return valid

    # Get properties of the tree we're placing
    props0 = tree_props[tree_index]

    # Get the z score of the tree we're placing
    z = props0[-1]
    # Compare to the z-score of the given pixel on the grid
    if z > z_score[i, j] + height_error:
        # We only accept z-scores below a given threshold
        return False

    # Tree grid dimensions
    n0 = tree_grid.shape[0]
    n1 = tree_grid.shape[1]

    # Get max radius of tree
    r = props0[2] * tree_scale
    # Check a local window around a tree to see if there's too much overlap with other trees
    w = ceil(r / dx) + 1
    for i1 in range(max(1, i - w), min(n0 - 1, i + w)):
        for j1 in range(max(1, j - w), min(n1 - 1, j + w)):
            if tree_grid[i1, j1] > 0:

                # Make sure tree is further than minimum distance
                dist = np.sqrt(((i - i1) * dx) ** 2 + ((j - j1) * dx) ** 2)
                if dist < min_radius:
                    return False

                # Get properties of this nearby tree
                props1 = tree_props[tree_grid[i1, j1] - 1]

                # Check for crown overlap
                collision = intersects(
                    dist,
                    props0,
                    props1,
                    tree_scale,
                    crown_profile_model=crown_profile_model,
                )
                if collision:
                    return False

    return True


@njit
def local_search(tree_grid, i, j, dx=1.0, max_dist=30.0):
    """
    Performs a local search to find an unoccupied pixel near a point.

    Parameters
    ----------
    tree_grid : ndarray
        Array of tree indexes of int64 type. Zero values indicate no tree. Positive values refer to a particular tree index.
    i : int
        Proposed row index for tree
    j : int
        Proposed column index for tree
    dx : float
        Pixel size used to compute distances.
    max_dist : float
        Determines how large of an area to search.

    Returns
    -------
    (int, int)
        Grid indices of free location or attempted indices.
    """

    # Tree grid dimensions
    n0 = tree_grid.shape[0]
    n1 = tree_grid.shape[1]

    # Check a local window around a tree to see if there's an open spot
    i2 = i
    j2 = j
    w = int(max_dist / dx) + 1
    for i1 in range(max(1, i - w), min(n0 - 1, i + w)):
        for j1 in range(max(1, j - w), min(n1 - 1, j + w)):
            i2 = i1
            j2 = j1

            if tree_grid[i1, j1] == 0:
                break

    return i2, j2


@njit
def sample_trees(
    chm: NDArray,
    z_score: NDArray,
    tree_grid: NDArray,
    tree_coords: NDArray,
    tree_props: NDArray,
    dx: float,
    tree_scale: NDArray,
    max_dist: NDArray,
    height_error: NDArray,
):
    """
    Perform Poisson disc sampling. On each iteration, try to place any trees not placed on previous iterations.
    Stringency of placement parameters are relaxed per iteration.

    Parameters
    ----------
    chm: NDArray
        Crown height model (CHM).
    z_score: NDArray
        A z_score for each CHM pixel representing its height compared to its neighbors.
    tree_grid: NDArray
        An array of the same size as the CHM used to mark tree locations.
    tree_coords : NDArray
        The initial pixel coordinates of all trees in a tree population.
    tree_props: NDArray
        The geometric parameters of each tree in a tree population.
    dx : float
        Size of a CHM pixel.
    tree_scale : NDArray
        Tree radii are multipled by this factor.
    max_dist: NDArray
        Max distance a tree can move from its original location.
    height_error: NDArray
        Used to control how stringent height based placement is.


    Returns
    -------
    tree_grid : NDArray
        The tree_grid with new tree locations marked.
    trees_to_place : list of ints
        The number of trees left to place per sampling iteration.
    """

    indexes = list(np.arange(len(tree_coords)))

    # This list will track the number of trees left to place at each iteration
    trees_to_place = []

    # Iterate until we've placed all trees
    for i in range(len(tree_scale)):
        new_indexes = []
        trees_to_place.append(len(indexes))

        for j in range(len(indexes)):
            tree_index = indexes[j]

            xi = tree_coords[tree_index, 0]
            xj = tree_coords[tree_index, 1]
            if i > 0:
                xi, xj = propose_index(
                    xi, xj, tree_grid.shape[0], tree_grid.shape[1], dx, dx, max_dist[i]
                )

            valid = accept(
                chm,
                z_score,
                tree_grid,
                tree_props,
                tree_index,
                xi,
                xj,
                dx,
                dx,
                tree_scale[i],
                height_error[i],
            )

            # If the tree's current position is good then place it on the grid
            if valid:
                tree_grid[xi, xj] = tree_index + 1
            else:
                # Otherwise add it to this list of indexes for subsequent passes
                new_indexes.append(tree_index)

        indexes = new_indexes
        if len(indexes) == 0:
            break

    # If there are some trees left over, just do a local search for an open space
    for j in range(len(indexes)):
        tree_index = indexes[j]
        xi = tree_coords[tree_index, 0]
        xj = tree_coords[tree_index, 1]

        xi, xj = local_search(tree_grid, xi, xj, dx=dx, max_dist=30.0)
        tree_grid[xi, xj] = tree_index + 1

    return tree_grid, trees_to_place



"""
This module contains functions for plotting data from the fastfuels_core
package.
"""

import numpy as np
import pyvista as pv
import matplotlib.pyplot as plt
import matplotlib.collections as collections


def plot_voxelized_tree(data, quantity="", **kwargs):
    viz_array = data.copy()
    viz_array[viz_array == 0] = -1
    grid = pv.UniformGrid()
    grid.dimensions = np.array(viz_array.shape) + 1
    grid.spacing = (1, 1, 1)
    grid.cell_data[quantity] = viz_array.flatten(order="F")
    grid = grid.threshold(0)
    # Set colormap max to 8

    pv.set_plot_theme("document")
    grid.plot(show_edges=False, show_grid=True, **kwargs)


def plot_tree_xy(treelist, return_fig_ax=False):
    fig, ax = plt.subplots()
    ax.scatter(treelist.data["X_m"], treelist.data["Y_m"], s=1)
    ax.set_aspect("equal")
    ax.autoscale_view()

    if return_fig_ax:
        return fig, ax
    else:
        plt.show()


def plot_tree_crowns(treelist, return_fig_ax=False):
    fig, ax = plt.subplots()
    patches = []
    face_colors = []
    for tree in treelist:
        circ = plt.Circle((tree.x, tree.y), tree.crown_radius)
        if tree.spcd == "202":
            face_colors.append("r")
        else:
            face_colors.append("g")
        patches.append(circ)
    plot_collection = collections.PatchCollection(
        patches, facecolors=face_colors, edgecolors="k", alpha=0.5
    )
    ax.add_collection(plot_collection)
    ax.autoscale_view()

    # Add red and green circles for legend
    ax.scatter([], [], c="r", alpha=0.5, label="Douglas-fir")
    ax.scatter([], [], c="g", alpha=0.5, label="Ponderosa pine")
    ax.legend()

    ax.set_xlabel("X (m)")
    ax.set_ylabel("Y (m)")

    if return_fig_ax:
        return fig, ax
    else:
        plt.show()


def get_canopy_profile(crown_radius, beta_params, n=14):
    """Returns a discretized canopy profile"""

    # beta parameters for ponderosa pine
    a, b, c = beta_params

    # Stirling's approximation for normalizing
    normalizer = np.sqrt(2 * np.pi) * (
        (a ** (a - 0.5) * b ** (b - 0.5)) / ((a + b) ** (a + b - 0.5))
    )

    z = np.linspace(0, 1, n)
    x = np.zeros_like(z)

    # discretized profile
    for i in range(n):
        x[i] = c * (((z[i] ** (a - 1)) * ((1 - z[i]) ** (b - 1)))) / normalizer

    # scale it up such that maximum point in x == crown radius
    factor = crown_radius / np.max(x)
    x *= factor
    z *= factor

    return x, z


def revolve_canopy(x, z):
    """Revolves the discrete profile points about the z axis"""

    n = x.shape[0]
    n_verts = n**2 - 3 * n + 4

    verts = np.zeros((n_verts, 3))
    theta = np.linspace(0, 2 * np.pi, n)

    verts[0, :] = [0, 0, 0]

    idx = 1
    for i in range(1, n - 1):
        for j in range(n - 1):
            verts[idx, 0] = x[i] * np.cos(theta[j])
            verts[idx, 1] = x[i] * np.sin(theta[j])
            verts[idx, 2] = z[i]
            idx += 1

    verts[-1, :] = [0, 0, z[-1]]

    return verts


def triangulate_verts(verts):
    """Triangulate the revolved point to a mesh"""

    n_verts = verts.shape[0]
    n = int((3 + np.sqrt(4 * n_verts - 7)) / 2)
    n_faces = 2 * n**2 - 6 * n + 4
    faces = np.zeros((n_faces, 4), dtype=int)

    idx = 0
    for i in range(1, n - 1):
        faces[idx] = [3, i, 0, i + 1]
        idx += 1
    faces[idx] = [3, n - 1, 0, 1]
    idx += 1

    for i in range(0, n - 3):
        j = i * (n - 1) + 1
        for k in range(j, j + n - 2):
            faces[idx] = [3, k, k + 1, k + n]
            idx += 1
            faces[idx] = [3, k, k + n, k + n - 1]
            idx += 1
        faces[idx] = [3, k + 1, j, k + 2]
        idx += 1
        faces[idx] = [3, k + 1, k + 2, k + n]
        idx += 1

    for i in range(n_verts - n, n_verts - 2):
        faces[idx] = [3, n_verts - 1, i, i + 1]
        idx += 1
    faces[idx] = [3, n_verts - 2, n_verts - n, n_verts - 1]

    return faces


def beta_canopy_verts_and_faces(tree, n=25):
    # def get_tree(xp, yp, zp, crown_radius, beta_params, n=25):
    """Returns nodes and faces of a beta canopy"""

    # 2D profile
    x, z = get_canopy_profile(
        tree.crown_radius,
        (tree.beta_canopy_a, tree.beta_canopy_b, tree.beta_canopy_c),
        n=n,
    )

    # now to 3D
    verts = revolve_canopy(x, z)

    # translate to (x,y,z)
    verts[:, 0] += tree.x
    verts[:, 1] += tree.y
    verts[:, 2] += tree.crown_base_height

    # create mesh
    faces = triangulate_verts(verts)

    return verts, faces


# def plot_beta_canopy_3d_plt(tree, ax=None, n=25):
#     """ Plots the canopy of a beta tree """
#
#     # get the nodes and faces
#     verts, faces = get_tree(tree, n=n)
#
#     # plot
#     if ax is None:
#         fig = plt.figure()
#         ax = fig.add_subplot(111, projection='3d')
#
#     ax.plot_trisurf(verts[:, 0], verts[:, 1], verts[:, 2], triangles=faces,
#                     color='green', alpha=0.5)
#
#     return ax


def get_tree_mesh(tree, n=25):
    verts, faces = beta_canopy_verts_and_faces(tree, n=n)
    mesh = pv.PolyData(verts, faces)
    return mesh


def plot_beta_canopy_3d_pyv(trees, n=25):
    # show the trees with pyvista (friendly VTK)
    pv.set_plot_theme("document")
    p = pv.Plotter()
    p.show_grid()
    try:
        for tree in trees:
            mesh = get_tree_mesh(tree, n=n)
            p.add_mesh(mesh, color="#00b500", smooth_shading=True)
            p.add_mesh(
                pv.Cylinder(
                    center=[tree.x, tree.y, (tree.height * (1 - tree.crown_ratio)) / 2],
                    direction=[0, 0, 1],
                    radius=tree.dbh / 200,
                    height=tree.height * (1 - tree.crown_ratio),
                ),
                color="#e5e7eb",
                smooth_shading=True,
            )
    except TypeError:
        tree = trees
        mesh = get_tree_mesh(trees, n=n)
        p.add_mesh(mesh, color="#00b500", smooth_shading=True)
        p.add_mesh(
            pv.Cylinder(
                center=[tree.x, tree.y, (tree.height * (1 - tree.crown_ratio)) / 2],
                direction=[0, 0, 1],
                radius=tree.dbh / 200,
                height=tree.height * (1 - tree.crown_ratio),
            ),
            color="#66493A",  # color="#e5e7eb",
            smooth_shading=True,
        )

    p.show()



# Core imports
from pathlib import Path

# Internal imports
from fastfuels_core.trees import (
    Tree,
    # BetaCrownProfile,
    JenkinsBiomassEquations,
    NSVBEquations,
    SPCD_PARAMS,
)

from fastfuels_core.crown_profile_models.beta import BetaCrownProfile

# External imports
import pytest
import numpy as np

TEST_PATH = Path(__file__).parent
TEST_DATA_PATH = TEST_PATH / "data"


class TestBetaCrownProfileModel:

    def test_get_radius_at_height(self):
        """
        Tests that the radius at a given height is calculated correctly.
        """
        model = BetaCrownProfile(
            species_code=122, crown_base_height=10, crown_length=10
        )

        # Test that the bottom of the crown and the top of the crown both have
        # a radius of 0
        assert model.get_radius_at_height(10) == 0.0
        assert model.get_radius_at_height(20) == 0.0

        # Test with a single float within the range of the tree's height
        assert model.get_radius_at_height(15) > 0

        # Test with a single float outside the range of the tree's height
        assert model.get_radius_at_height(-1) == 0
        assert model.get_radius_at_height(21) == 0

        # Test with a numpy array of floats, all within the range of the tree's height
        height_values = np.array([12, 14, 16, 18])
        radii = model.get_radius_at_height(height_values)
        assert len(radii) == len(height_values)
        assert all(radii > 0)

        # Test with a numpy array of floats, some within and some outside the range of the tree's height
        height_values = np.array([-1, 10, 15, 20, 21])
        radii = model.get_radius_at_height(height_values)
        assert len(radii) == len(height_values)
        assert radii[0] == 0
        assert radii[1] == 0
        assert radii[2] > 0
        assert radii[3] == 0
        assert radii[4] == 0

    def test_get_normalized_height(self):
        """
        Tests that the normalized height is calculated correctly.
        """
        # Suppose that I have a tree with a height of 20m, a crown base height
        # of 10m, and a crown length of 10m
        model = BetaCrownProfile(
            species_code=122, crown_base_height=10, crown_length=10
        )
        assert model._get_normalized_height(0) < 0
        assert model._get_normalized_height(5) < 0
        assert model._get_normalized_height(10) == 0
        assert model._get_normalized_height(15) == 0.5
        assert model._get_normalized_height(20) == 1
        assert model._get_normalized_height(25) > 1

        # Suppose that I have a tree with a height of 100m, a crown base height
        # of 20m, and a crown length of 80m
        model = BetaCrownProfile(
            species_code=122, crown_base_height=20, crown_length=80
        )
        assert model._get_normalized_height(0) < 0
        assert model._get_normalized_height(10) < 0
        assert model._get_normalized_height(20) == 0
        assert model._get_normalized_height(40) == 0.25
        assert model._get_normalized_height(60) == 0.5
        assert model._get_normalized_height(80) == 0.75
        assert model._get_normalized_height(100) == 1
        assert model._get_normalized_height(120) > 1

        # Test with numpy
        heights = np.linspace(10, 20, 100)
        model = BetaCrownProfile(
            species_code=122, crown_length=10, crown_base_height=10
        )
        normalized_heights = model._get_normalized_height(heights)
        assert len(normalized_heights) == 100
        assert normalized_heights[0] == 0
        assert normalized_heights[-1] == 1.0
        for i in range(1, 100):
            assert normalized_heights[i] > normalized_heights[i - 1]

    def test_get_radius_at_normalized_height(self):
        """
        Tests that the radius at a given normalized height is calculated correctly.
        """
        model = BetaCrownProfile(
            species_code=122, crown_base_height=10, crown_length=10
        )

        # Test that the bottom of the crown and the top of the crown both have
        # a radius of 0
        assert model._get_radius_at_normalized_height(0) == 0.0
        assert model._get_radius_at_normalized_height(1) == 0.0

        # Test with a single float within the range [0, 1]
        assert model._get_radius_at_normalized_height(0.5) > 0

        # Test with a single float outside the range [0, 1]
        assert model._get_radius_at_normalized_height(-1) == 0.0
        assert model._get_radius_at_normalized_height(2) == 0.0

        # Test with a numpy array of floats, all within the range [0, 1]
        z_values = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
        radii = model._get_radius_at_normalized_height(z_values)
        assert len(radii) == len(z_values)
        assert all(radii > 0)

        # Test with a numpy array of floats, some within and some outside the range [0, 1]
        z_values = np.array([-1, 0, 0.5, 1, 2])
        radii = model._get_radius_at_normalized_height(z_values)
        assert len(radii) == len(z_values)
        assert radii[0] == 0.0
        assert radii[1] == 0.0
        assert radii[2] > 0
        assert radii[3] == 0.0
        assert radii[4] == 0.0

    def test_get_max_radius(self):
        """
        Tests that the maximum radius of the crown is calculated correctly.
        """
        # Test with species group 4, crown base height 10, and crown length 10
        model = BetaCrownProfile(
            species_code=122, crown_base_height=10, crown_length=10
        )
        max_crown_radius = model.get_max_radius()
        assert max_crown_radius > 0
        heights = np.linspace(10, 20, 1000)
        radii = model.get_radius_at_height(heights)
        assert np.all(max_crown_radius >= radii)

        # Test with species group 1, crown base height 5, and crown length 15
        model = BetaCrownProfile(species_code=202, crown_base_height=5, crown_length=15)
        max_crown_radius = model.get_max_radius()
        assert max_crown_radius > 0
        heights = np.linspace(5, 20, 1000)
        radii = model.get_radius_at_height(heights)
        assert np.all(max_crown_radius >= radii)

        # Test with species group 2, crown base height 20, and crown length 5
        model = BetaCrownProfile(species_code=747, crown_base_height=20, crown_length=5)
        max_crown_radius = model.get_max_radius()
        assert max_crown_radius > 0
        heights = np.linspace(20, 25, 1000)
        radii = model.get_radius_at_height(heights)
        assert np.all(max_crown_radius >= radii)


class TestJenkinsBiomassEquations:
    def test_init(self):
        # Test that the class can be initialized
        model = JenkinsBiomassEquations(species_code=122, diameter=24)
        assert model._species_group == 4
        assert model._is_softwood == 1
        assert np.isclose(model._sapling_adjustment, 0.43373999999999996)

        # Test with a different species code
        model = JenkinsBiomassEquations(species_code=989, diameter=100)
        assert model._species_group == 8
        assert model._is_softwood == 0
        assert model._sapling_adjustment == 0.84031

        # Test with an invalid species code
        with pytest.raises(ValueError):
            JenkinsBiomassEquations(species_code=32480910, diameter=100)

        # Test with an invalid diameter
        with pytest.raises(ValueError):
            JenkinsBiomassEquations(species_code=122, diameter=-100)

        # Test initializing all species codes
        for species_code in SPCD_PARAMS:
            JenkinsBiomassEquations(species_code=int(species_code), diameter=100)

    def test_estimate_above_ground_biomass(self):
        # Test with a known species code and diameter
        model = JenkinsBiomassEquations(species_code=122, diameter=24)
        biomass = model._estimate_above_ground_biomass()
        assert biomass > 0, "Biomass should be a positive number"

        # Test each species code with a random diameter
        for species_code in SPCD_PARAMS:
            diameter = np.random.uniform(0, 100)
            model = JenkinsBiomassEquations(
                species_code=int(species_code), diameter=diameter
            )
            if model._species_group == -1:
                continue
            biomass = model._estimate_above_ground_biomass()
            assert (
                biomass > 0
            ), f"{species_code} and {diameter} should yield a positive biomass"

    def test_estimate_foliage_biomass(self):
        # Test with a known species code and diameter
        model = JenkinsBiomassEquations(species_code=122, diameter=24)
        biomass = model.estimate_foliage_biomass()
        assert biomass > 0, "Biomass should be a positive number"

        # Test each species code with a random diameter
        for species_code in SPCD_PARAMS:
            diameter = np.random.uniform(0, 100)
            model = JenkinsBiomassEquations(
                species_code=int(species_code), diameter=diameter
            )
            if model._species_group == -1:
                continue
            biomass = model.estimate_foliage_biomass()
            assert (
                biomass > 0
            ), f"{species_code} and {diameter} should yield a positive biomass"


class TestNSVBEquations:
    def test_init(self):
        # Test that the class can be initialized
        NSVBEquations(species_code=122, diameter=24, height=20)

        # Test with a different species code
        NSVBEquations(species_code=989, diameter=100, height=20)

        # Test with an invalid species code
        with pytest.raises(ValueError):
            NSVBEquations(species_code=32480910, diameter=100, height=20)

        # Test with an invalid diameter
        with pytest.raises(ValueError):
            NSVBEquations(species_code=122, diameter=-100, height=20)

    def test_estimate_foliage_biomass_example_1(self):
        # Compare to Example 1 of the GTR
        dia_in = 20
        dia_cm = dia_in * 2.54
        ht_ft = 110
        ht_m = ht_ft * 0.3048
        division = "240"
        model = NSVBEquations(
            species_code=202, diameter=dia_cm, height=ht_m, division=division
        )
        foliage_biomass = model.estimate_foliage_biomass()

        foliage_biomass_lb = foliage_biomass * 2.20462
        assert np.isclose(foliage_biomass_lb, 83.63478892024017)

    def test_estimate_foliage_biomass_example_2(self):
        # Compare to Example 2 of the GTR
        spcd = 316
        dia_in = 11.1
        dia_cm = dia_in * 2.54
        ht_ft = 38
        ht_m = ht_ft * 0.3048
        division = "M210"
        model = NSVBEquations(
            species_code=spcd, diameter=dia_cm, height=ht_m, division=division
        )
        foliage_biomass = model.estimate_foliage_biomass()

        foliage_biomass_lb = foliage_biomass * 2.20462
        assert np.isclose(foliage_biomass_lb, 22.807960563788)


def test_tree():
    tree = Tree(species_code=122, status_code=1, diameter=5, height=20, crown_ratio=0.5)
    assert isinstance(tree.biomass_allometry_model, NSVBEquations)
